import * as THREE from 'three';
import { useEffect, useRef, useState } from 'react';
import { retrieveGLTF, storeGLTF } from '../../../../../utils/indexDB/idbUtils';
import { GLTF, GLTFLoader } from 'three/examples/jsm/loaders/GLTFLoader';
import { DRACOLoader } from 'three/examples/jsm/loaders/DRACOLoader';
import { useFrame, useThree } from '@react-three/fiber';
import { useActiveTool, useDeletableFloorItem, useRenderDistance, useSelectedFloorItem } from '../../../../../store/builder/store';
import { AssetBoundingBox } from './assetBoundingBox';
import { CameraControls } from '@react-three/drei';
import { useAssetsStore } from '../../../../../store/builder/useAssetStore';
import { useEventsStore } from "../../../../../store/simulation/useEventsStore";
import { useProductStore } from "../../../../../store/simulation/useProductStore";
import { useSocketStore } from '../../../../../store/builder/store';

function Model({ asset }: { asset: Asset }) {
    const { camera, controls } = useThree();
    const { activeTool } = useActiveTool();
    const { removeAsset } = useAssetsStore();
    const { socket } = useSocketStore();
    const { deletableFloorItem, setDeletableFloorItem } = useDeletableFloorItem();
    const { setSelectedFloorItem } = useSelectedFloorItem();
    const { renderDistance } = useRenderDistance();
    const [isRendered, setIsRendered] = useState(false);
    const url_Backend_dwinzo = `http://${process.env.REACT_APP_SERVER_MARKETPLACE_URL}`;
    const [gltfScene, setGltfScene] = useState<GLTF | null>(null);
    const [boundingBox, setBoundingBox] = useState<THREE.Box3 | null>(null);
    const groupRef = useRef<THREE.Group>(null);

    useEffect(() => {
        const loader = new GLTFLoader();
        const dracoLoader = new DRACOLoader();

        dracoLoader.setDecoderPath('https://cdn.jsdelivr.net/npm/three@0.160.0/examples/jsm/libs/draco/gltf/');
        loader.setDRACOLoader(dracoLoader);
        const loadModel = async () => {
            try {
                // Check Cache
                const cachedModel = THREE.Cache.get(asset.assetId!);
                if (cachedModel) {
                    setGltfScene(cachedModel);
                    calculateBoundingBox(cachedModel.scene);
                    return;
                }

                // Check IndexedDB
                const indexedDBModel = await retrieveGLTF(asset.assetId!);
                if (indexedDBModel) {
                    const blobUrl = URL.createObjectURL(indexedDBModel);
                    loader.load(blobUrl, (gltf) => {
                        URL.revokeObjectURL(blobUrl);
                        THREE.Cache.remove(blobUrl);
                        THREE.Cache.add(asset.assetId!, gltf);
                        setGltfScene(gltf);
                        calculateBoundingBox(gltf.scene);
                    },
                        undefined,
                        (error) => {
                            echo.error(`[IndexedDB] Error loading ${asset.modelName}:`);
                            URL.revokeObjectURL(blobUrl);
                        }
                    );
                    return;
                }

                // Fetch from Backend
                const modelUrl = `${url_Backend_dwinzo}/api/v2/AssetFile/${asset.assetId!}`;
                loader.load(modelUrl, async (gltf) => {
                    const modelBlob = await fetch(modelUrl).then((res) => res.blob());
                    await storeGLTF(asset.assetId!, modelBlob);
                    THREE.Cache.add(asset.assetId!, gltf);
                    setGltfScene(gltf);
                    calculateBoundingBox(gltf.scene);
                },
                    undefined,
                    (error) => {
                        echo.error(`[Backend] Error loading ${asset.modelName}:`);
                    }
                );
            } catch (err) {
                console.error("Failed to load model:", asset.assetId, err);
            }
        };

        const calculateBoundingBox = (scene: THREE.Object3D) => {
            const box = new THREE.Box3().setFromObject(scene);
            setBoundingBox(box);
        };

        loadModel();

    }, []);

    useFrame(() => {
        const assetPosition = new THREE.Vector3(...asset.position);
        if (!isRendered && assetPosition.distanceTo(camera.position) <= renderDistance) {
            setIsRendered(true);
        } else if (isRendered && assetPosition.distanceTo(camera.position) > renderDistance) {
            setIsRendered(false);
        }
    })

    const handleDblClick = (asset: Asset) => {
        if (asset) {
            if (activeTool === "cursor" && boundingBox && groupRef.current) {
                const size = boundingBox.getSize(new THREE.Vector3());
                const center = boundingBox.getCenter(new THREE.Vector3());

                const front = new THREE.Vector3(0, 0, 1);
                groupRef.current.localToWorld(front);
                front.sub(groupRef.current.position).normalize();

                const distance = Math.max(size.x, size.y, size.z) * 2;
                const newPosition = center.clone().addScaledVector(front, distance);

                (controls as CameraControls).setPosition(
                    newPosition.x,
                    newPosition.y,
                    newPosition.z,
                    true
                );
                (controls as CameraControls).setTarget(center.x, center.y, center.z, true);
                (controls as CameraControls).fitToBox(groupRef.current!, true, {
                    cover: true,
                    paddingTop: 5,
                    paddingLeft: 5,
                    paddingBottom: 5,
                    paddingRight: 5,
                });
                setSelectedFloorItem(groupRef.current);

            }
        }
    };

    const handleClick = (asset: Asset) => {
        if (activeTool === 'delete' && deletableFloorItem && deletableFloorItem.uuid === asset.modelUuid) {
            const email = localStorage.getItem('email')
            const organization = (email!.split("@")[1]).split(".")[0];


            //REST

            // const response = await deleteFloorItem(organization, asset.modelUuid, asset.modelName);

            //SOCKET 

            const data = {
                organization: organization,
                modelUuid: asset.modelUuid,
                modelName: asset.modelName,
                socketId: socket.id
            }

            const response = socket.emit('v2:model-asset:delete', data)

            useEventsStore.getState().removeEvent(asset.modelUuid);
            useProductStore.getState().deleteEvent(asset.modelUuid);

            if (response) {

                const storedItems = JSON.parse(localStorage.getItem("FloorItems") || '[]');
                const updatedStoredItems = storedItems.filter((item: { modelUuid: string }) => item.modelUuid !== asset.modelUuid);
                localStorage.setItem("FloorItems", JSON.stringify(updatedStoredItems));

                removeAsset(asset.modelUuid);

                echo.success("Model Removed!");
            }

        }
    }

    const handlePointerOver = (asset: Asset) => {
        if (activeTool === "delete") {
            if (deletableFloorItem && deletableFloorItem.uuid === asset.modelUuid) {
                return;
            } else {
                setDeletableFloorItem(groupRef.current);
            }
        }
    }

    const handlePointerOut = (asset: Asset) => {
        if (activeTool === "delete" && deletableFloorItem && deletableFloorItem.uuid === asset.modelUuid) {
            setDeletableFloorItem(null);
        }
    }

    return (
        <group
            name='Asset Model'
            ref={groupRef}
            uuid={asset.modelUuid}
            position={asset.position}
            rotation={asset.rotation}
            visible={asset.isVisible}
            userData={asset}
            onDoubleClick={(e) => {
                e.stopPropagation();
                handleDblClick(asset);
            }}
            onClick={(e) => {
                e.stopPropagation();
                handleClick(asset);
            }}
            onPointerOver={(e) => {
                e.stopPropagation();
                handlePointerOver(asset);
            }}
            onPointerOut={(e) => {
                e.stopPropagation();
                handlePointerOut(asset);
            }}
        >
            {gltfScene && (
                isRendered ? (
                    <primitive object={gltfScene.scene.clone()} />
                ) : (
                    <AssetBoundingBox asset={asset} boundingBox={boundingBox} />
                )
            )}
        </group>
    );
}

export default Model;